/**************************************************************************
* Copyright (c) 2006, 2007 Dusan Stastny. All rights reserved.            *
*                                                                         *
* This software is free software; you can redistribute it and/or modify   *
* it under the terms of the GNU General Public License as published by	  *
* the Free Software Foundation; either version 2 of the License, or       *
* (at your option) any later version.                                     *
*                                                                         *
* This software is distributed in the hope that it will be useful, but    *
* WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        *
* General Public License for more details.                                *
*                                                                         *
* You should have received a copy of the GNU General Public License       *
* along with DS Java Disassembler; see the file COPYING. If not, write    *
* to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,      *
* Boston, MA 02111-1307 USA.                                              *
*                                                                         *
*   Dusan Stastny                                                         *
*   Jizni Svahy 804              info@developservice.cz                   *
*   330 27 Vejprnice             http://www.developservice.cz             *
*   Czech Republic - EUROPE                                               *
**************************************************************************/

/** @file
   The DS Java Disassembler disassembles Java .CLASS files to stdout.
   The output intended for analysis and understanding Java byte code, not recompilation. 
   <PRE>
   Using: jdis [-b] source

       -b        not expand attributes
       source    .CLASS file
   </PRE>
   @author Dusan Stastny
*/

#include "jdis.h"
#include <stdio.h>

/* expand attributes? */
int exp_atr = 1;

/* constant pool tags */
char *cp_tags[] = {"", "Utf8", "", "Integer", "Float", "Long", "Double", "Class", 
	"String", "FieldRef", "MethodRef", "InterfaceMethodRef", "Name&Type"};

const char digits[] = "0123456789";


/**
   Converts an encoded UTF-8 string into an Unicode string.
   @param str a pointer to the output buffer to receive converted Unicode string,
               returned Unicode string will be terminated by zero character
   @param utf8 an UTF-8 string to convert
   @return number of the converted Unicode characters in output buffer or negative value if an error occurs
*/
int fromUtf8(jchar *str, const char *utf8) {
	WCHAR *buf = str;
	int i = 0, lbuf = 0, err = 0;
	while (utf8[i]) {
		if (i >= 65535) { err = -1; break; }
        if ((utf8[i] & 0x80) == 0)
			buf[lbuf++] = (WCHAR)(utf8[i++] & 0xFF);
		else 
			if ((utf8[i] & 0xE0) == 0xC0) {
				WCHAR ch1 = (WCHAR)(unsigned char)utf8[i++], ch2 = (WCHAR)(unsigned char)utf8[i++];
				if ((ch2 & 0xC0) != 0x80) {	err = -1; break; }
				buf[lbuf++] = (WCHAR)(((ch1 & 0x1F) << 6) | (ch2 & 0x3F));
			} else 
				if ((utf8[i] & 0xF0) == 0xE0) {
					WCHAR ch1 = (WCHAR)(unsigned char)utf8[i++], ch2 = (WCHAR)(unsigned char)utf8[i++], ch3 = (WCHAR)(unsigned char)utf8[i++];
					if (((ch2 & 0xC0) != 0x80) || ((ch3 & 0xC0) != 0x80)) {	err = -1; break; }
					buf[lbuf++] = (WCHAR)(((ch1 & 0x0F) << 12) | ((ch2 & 0x3F) << 6) | (ch3 & 0x3F));
				} else {
					err = -1; break; 
				}
	}
	buf[lbuf] = 0;
	return err ? err : lbuf;
}


/**
   Loads a class definition from the raw class data.
   <B>ATTENTION:</B> All pointers to another class definitions in result
   point to (UTF-8) class names and must be replaced with pointers to correct class definitions!
   @param buf the raw class data
   @param bufLen length of the raw class data
   @return loaded class definition or NULL if an error occurs
*/
clsDefinition *vmLoadClass(const jbyte *buf, jsize bufLen) {
	clsDefinition cls, *ret = NULL; char *s;
	jint i, j, k, cnt; jbyte *p; const jbyte *pbuf; 
	const jbyte *pool;		/* pointer to buffer */
	const jbyte *flds;		/* pointer to buffer */
	const jbyte *mets;		/* pointer to buffer */
	const jbyte *atrs;		/* pointer to buffer */
	jint pool_len = 0;		/* needed size of memory in bytes */
	jint atrnum = 0;		/* total number of the attributes */
	jint atrsize = 0;		/* total size of the attributes */
	if (!buf || (bufLen <= 0)) return NULL;
	vmMemSetZero(&cls, sizeof(cls));
	
	/* are the .class file data? */
	if (getClassInt(buf) != 0xCAFEBABE) return NULL;
	cls.minor_version = getClassUShort(buf + 4);
	cls.major_version = getClassUShort(buf + 6);
	cls.constant_pool_count = getClassUShort(buf + 8);
	
	/* compute needed memory size */
	pool = buf + 10; pbuf = pool;
	for (i = 1; i < cls.constant_pool_count; i++)
		switch (*(pbuf++)) {
			case CONSTANT_Utf8: {
				jint len = getClassUShort(pbuf);
				pool_len += len + 1; pbuf += len + 2;
				break; }
			case CONSTANT_Class: 
			case CONSTANT_String: 
				pbuf += 2; break;
			case CONSTANT_Integer: 
			case CONSTANT_Float: 
			case CONSTANT_NameAndType: 
				pbuf += 4; break;
			case CONSTANT_FieldRef:
				pool_len += sizeof(jfieldID); pbuf += 4; break;
			case CONSTANT_MethodRef: 
			case CONSTANT_InterfaceMethodRef:
				pool_len += sizeof(jmethodID); pbuf += 4; break;
			case CONSTANT_Long: 
			case CONSTANT_Double: 
				i++; pbuf += 8; break;
			default: 
				return NULL;	/* unknown */
		};
	cls.access_flags = ACC_FREEMEM | getClassUShort(pbuf); pbuf += 6;
	cls.interfaces_count = getClassUShort(pbuf); pbuf += 2 + 2 * cls.interfaces_count;
	cnt = getClassUShort(pbuf); pbuf += 2; flds = pbuf;
	for (i = 0; i < cnt; i++) {
		if (getClassUShort(pbuf) & ACC_STATIC) cls.class_fields_count++; 
			else cls.instance_fields_count++;
		j = getClassUShort(pbuf + 6); /* attributes count */
		pbuf += 8; atrnum += j;
		for (k = 0; k < j; k++) {
			jint sz = getClassInt(pbuf + 2); pbuf += 6 + sz; atrsize += sz;
		}
	}
	cnt = getClassUShort(pbuf); pbuf += 2; mets = pbuf;
	for (i = 0; i < cnt; i++) {
		if (getClassUShort(pbuf) & ACC_STATIC) cls.class_methods_count++; 
			else cls.instance_methods_count++;
		j = getClassUShort(pbuf + 6); /* attributes count */
		pbuf += 8; atrnum += j;
		for (k = 0; k < j; k++) {
			jint sz = getClassInt(pbuf + 2); pbuf += 6 + sz; atrsize += sz;
		}
	}
	cls.attributes_count = getClassUShort(pbuf); pbuf += 2; atrs = pbuf;
	atrnum += cls.attributes_count;
	for (k = 0; k < cls.attributes_count; k++) {
		jint sz = getClassInt(pbuf + 2); pbuf += 6 + sz; atrsize += sz; 
	}
	
	/* create new class definition */
	i = sizeof(clsDefinition) + 
		cls.constant_pool_count * sizeof(clsConstPool) + pool_len +
		cls.interfaces_count * sizeof(clsDefinition*) + 
		(cls.class_fields_count + cls.instance_fields_count) * sizeof(clsField) + 
		(cls.class_methods_count + cls.instance_methods_count) * (sizeof(clsMethod*) + sizeof(clsMethod)) + 
		atrnum * sizeof(clsAttribute) + atrsize;
	p = (jbyte*)ret = vmMemAlloc(i);
	if (p == NULL) return NULL;
	vmMemSetZero(ret, i);
	vmMemCpy(ret, &cls, sizeof(cls));
	p += sizeof(clsDefinition);
	
	/* load the constant pool */
	ret->constant_pool = (clsConstPool*)p;
	p += cls.constant_pool_count * sizeof(clsConstPool);
	pbuf = pool;
	for (i = 1; i < cls.constant_pool_count; i++)
		switch (ret->constant_pool[i].tag = *(pbuf++)) {
			case CONSTANT_Utf8: {
				jint len = ret->constant_pool[i].i = getClassUShort(pbuf);
				vmMemCpy(ret->constant_pool[i].p = p, pbuf + 2, len); 
				p += len + 1; pbuf += len + 2;
				break; }
			case CONSTANT_Class:
			case CONSTANT_String: 
				ret->constant_pool[i].i = getClassUShort(pbuf);
				pbuf += 2; break;
			case CONSTANT_Integer:
			case CONSTANT_Float:
				ret->constant_pool[i].p = &(ret->constant_pool[i].i);
				ret->constant_pool[i].i = getClassInt(pbuf);
				pbuf += 4; break;
			case CONSTANT_NameAndType: 
NAMETYPE:		ret->constant_pool[i].i = getClassUShort(pbuf);
				ret->constant_pool[i].j = getClassUShort(pbuf + 2);
				pbuf += 4; break;
			case CONSTANT_FieldRef:
				ret->constant_pool[i].p = p; p += sizeof(jfieldID);
				goto NAMETYPE;
			case CONSTANT_MethodRef: 
			case CONSTANT_InterfaceMethodRef:
				ret->constant_pool[i].p = p; p += sizeof(jmethodID);
				goto NAMETYPE;
			case CONSTANT_Long: 
			case CONSTANT_Double: 
				ret->constant_pool[i].p = &(ret->constant_pool[i].i);
				ret->constant_pool[i].i = getClassInt(pbuf + 4);
				ret->constant_pool[i].j = getClassInt(pbuf);
				i++;
				ret->constant_pool[i].tag = CONSTANT_Integer;
				ret->constant_pool[i].p = &(ret->constant_pool[i].i);
				ret->constant_pool[i].i = getClassInt(pbuf);
				pbuf += 8; break;
		};
	for (i = 1; i < cls.constant_pool_count; i++)
		switch (ret->constant_pool[i].tag) {
			case CONSTANT_Class:
			case CONSTANT_String:
				ret->constant_pool[i].p = ret->constant_pool[ret->constant_pool[i].i].p;
				break;
			case CONSTANT_NameAndType:
				ret->constant_pool[i].p = ret->constant_pool[ret->constant_pool[i].i].p;
				ret->constant_pool[i].q = ret->constant_pool[ret->constant_pool[i].j].p;
				break;
		};

	/* class name */
	pbuf += 2;
	ret->name = ret->constant_pool[getClassUShort(pbuf)].p; pbuf += 2;
	ret->super_class = ret->constant_pool[getClassUShort(pbuf)].p; pbuf += 4;

	/* interfaces */
	if (ret->interfaces_count > 0) {
		ret->interfaces = (clsDefinition**)p;
		p += ret->interfaces_count * sizeof(clsDefinition*);
	}
	for (i = 0; i < ret->interfaces_count; i++) {
		ret->interfaces[i] = (clsDefinition*)getClassUShort(pbuf); pbuf += 2;
	}

	/* fields */
	pbuf = flds;
	cnt = cls.class_fields_count + cls.instance_fields_count;
	ret->class_fields_count = ret->instance_fields_count = 0;
	if (cls.class_fields_count > 0) ret->class_fields = (clsField*)p; p += cls.class_fields_count * sizeof(clsField);
	if (cls.instance_fields_count > 0) ret->instance_fields = (clsField*)p; p += cls.instance_fields_count * sizeof(clsField);
	for (i = 0; i < cnt; i++) {
		clsField *fld;
		jushort flags = getClassUShort(pbuf); 
		if (flags & ACC_STATIC)
			fld = ret->class_fields + ret->class_fields_count++;
		else
			fld = ret->instance_fields + ret->instance_fields_count++;
		fld->access_flags = flags; fld->clazz = ret;
		fld->name = ret->constant_pool[getClassUShort(pbuf + 2)].p;
		fld->descriptor = ret->constant_pool[getClassUShort(pbuf + 4)].p;
		/* attributes */
		fld->attributes_count = getClassUShort(pbuf + 6); pbuf += 8;
		fld->attributes = (clsAttribute*)p; p += fld->attributes_count * sizeof(clsAttribute);
		for (k = 0; k < fld->attributes_count; k++) {
			clsAttribute *atr = fld->attributes + k;
			atr->name = ret->constant_pool[getClassUShort(pbuf)].p; 
			atr->len = getClassInt(pbuf + 2); atr->data = p; p += atr->len; 
			vmMemCpy(atr->data, pbuf + 6, atr->len); pbuf += 6 + atr->len;
		}
	}

	/* methods */
	pbuf = mets;
	cnt = cls.class_methods_count + cls.instance_methods_count;
	ret->class_methods_count = ret->instance_methods_count = 0;
	if (cls.class_methods_count > 0) ret->class_methods = (clsMethod**)p; p += cls.class_methods_count * sizeof(clsMethod*);
	if (cls.instance_methods_count > 0) ret->instance_methods = (clsMethod**)p; p += cls.instance_methods_count * sizeof(clsMethod*);
	for (i = 0; i < cnt; i++) {
		clsMethod *met = (clsMethod*)p; p += sizeof(clsMethod);
		met->clazz = ret;
		met->access_flags = getClassUShort(pbuf);
		if (met->access_flags & ACC_STATIC) 
			ret->class_methods[ret->class_methods_count++] = met;
		else 
			ret->instance_methods[ret->instance_methods_count++] = met;
		met->name = ret->constant_pool[getClassUShort(pbuf + 2)].p;
		s = met->descriptor = ret->constant_pool[getClassUShort(pbuf + 4)].p;
		s++;
		while (*s && (*s != ')')) {
			met->parameters_count++;
			if ((*s == 'L') || (*s == '[')) while (*s && (*s != ';')) s++;
			s++;
		}
		met->stack_size = met->parameters_count + 1;
		if (*s == ')') met->return_type = s[1];
		/* attributes */
		met->attributes_count = getClassUShort(pbuf + 6); pbuf += 8;
		met->attributes = (clsAttribute*)p; p += met->attributes_count * sizeof(clsAttribute);
		for (k = 0; k < met->attributes_count; k++) {
			clsAttribute *atr = met->attributes + k;
			atr->name = ret->constant_pool[getClassUShort(pbuf)].p; 
			atr->len = getClassInt(pbuf + 2); atr->data = p; p += atr->len; 
			vmMemCpy(atr->data, pbuf + 6, atr->len); pbuf += 6 + atr->len;
			if (!vmStrCmpUTF(atr->name, "Code")) met->code = atr->data;
		}
	}

	/* attributes */
	pbuf = atrs;
	if (ret->attributes_count > 0) ret->attributes = (clsAttribute*)p; 
	p += ret->attributes_count * sizeof(clsAttribute);
	for (k = 0; k < ret->attributes_count; k++) {
		clsAttribute *atr = ret->attributes + k;
		atr->name = ret->constant_pool[getClassUShort(pbuf)].p; 
		atr->len = getClassInt(pbuf + 2); atr->data = p; p += atr->len; 
		vmMemCpy(atr->data, pbuf + 6, atr->len); pbuf += 6 + atr->len;
	}

	return ret;
}


/* Gets long integer. */
char *getLong(jlong num) {
	static char buf[64];
	int i = 62; buf[i--] = 0;			
	if (num < 0) { printf("-"); num = -num; }
	do {
		buf[i--] = digits[num % (jlong)10];
		num = num / (jlong)10;
	} while (num && i);
	return buf + i + 1;
}


/* Prints UTF-8 string. */
void printUtf8(const char *str, jint len) {
	jchar *buf;
	buf = vmMemAlloc(2 * len + 4);			
	if (buf != NULL) {
		/* UTF-8 to Unicode conversion */
		fromUtf8(buf, str);
		vmPrintf(L"%s", buf);
		vmMemFree(buf);
	} else
		printf("%s", str);
}


/* Gets type. */
char *getType(const char *d) {
	static char buf[4096];
	int i, dim;
	buf[0] = 0;
	dim = 0; while (*d && (*d == '[')) { dim++; d++; }	
	switch (*d) {
		case 'V': strcat(buf, "void"); break;
		case 'Z': strcat(buf, "boolean"); break;
		case 'B': strcat(buf, "byte"); break;
		case 'C': strcat(buf, "char"); break;
		case 'S': strcat(buf, "short"); break;
		case 'I': strcat(buf, "int"); break;
		case 'J': strcat(buf, "long"); break;
		case 'F': strcat(buf, "float"); break;
		case 'D': strcat(buf, "double"); break;
		default: {
			const char *p; 
			p = (*d == 'L') ? d + 1 : d;
			while (*p && (*p != ';')) {
				char pom[] = ".";
				if (*p != '/') pom[0] = *p;
				strcat(buf, pom);
				p++;
			}
			break;
		}
	}	
	for (i = 0; i < dim; i++) strcat(buf, "[]");
	return buf;
}


/* Gets constant string. */
char* getConstantString(clsDefinition *cls, jint idx) {
	static char buf[4096], *s;
	clsConstPool *c = cls->constant_pool + idx;
	if ((idx > 0) && (idx < cls->constant_pool_count))
		switch (cls->constant_pool[idx].tag) {
			case CONSTANT_Integer:
				sprintf(buf, "{INTEGER: %d}", (int)*(jint*)(c->p));
				return buf;
			case CONSTANT_Long:
				sprintf(buf, "{LONG: %s}", getLong(*(jlong*)(c->p)));
				return buf;
			case CONSTANT_Float:
				sprintf(buf, "{FLOAT: %g}", (double)*(jfloat*)(c->p));
				return buf;
			case CONSTANT_Double:
				sprintf(buf, "{DOUBLE: %g}", (double)*(jdouble*)(c->p));
				return buf;
			case CONSTANT_Utf8:
				return cls->constant_pool[idx].p;
			case CONSTANT_String:
				sprintf(buf, "{STRING: \"%s\"}", c->p);
				return buf;
			case CONSTANT_Class:
				sprintf(buf, "{CLASS: %s}", c->p);
				s = buf;
				while (*s) { if (*s == '/') *s = '.'; s++; }
				return buf;
			case CONSTANT_FieldRef:
				sprintf(buf, "{FIELD: %s %s.%s}",
					getType(cls->constant_pool[c->j].q),
					cls->constant_pool[c->i].p, 
					cls->constant_pool[c->j].p);
				s = buf;
				while (*s) { if (*s == '/') *s = '.'; s++; }
				return buf;
			case CONSTANT_MethodRef:
			case CONSTANT_InterfaceMethodRef:
				s = cls->constant_pool[c->j].q;
				while (*s && (*s != ')')) s++; if (*s == ')') s++;
				sprintf(buf, "{METHOD: %s %s.%s(", getType(s), 
					cls->constant_pool[c->i].p, cls->constant_pool[c->j].p);
				s = cls->constant_pool[c->j].q;
				while (*s && (*s != '(')) s++;
				while (*s && (*s != ')')) {
					if (*s == '(') s++; else vmStrCatUTF(buf, ", ");
					if (*s == ')') break;
					sprintf(buf + vmStrLenUTF(buf), getType(s));
					while (*s && (*s == '[')) s++;
					if (*s == 'L') {
						while (*s && (*s != ';')) s++;
						if (*s == ';') s++;
					} else
						if (*s) s++;
				}
				vmStrCatUTF(buf, ")}");
				s = buf;
				while (*s) { if (*s == '/') *s = '.'; s++; }
				return buf;
		}
	return "";
}


/* Prints class flags. */
void printClassFlags(jint flags) {
	int cnt = 0;
	if (flags & ACC_PUBLIC) printf("%spublic", cnt++ ? " " : "");
	if (flags & ACC_FINAL) printf("%sfinal", cnt++ ? " " : "");
	if (flags & ACC_SUPER) printf("%ssuper", cnt++ ? " " : "");
	if (flags & ACC_INTERFACE) printf("%sinterface", cnt++ ? " " : "");
	if (flags & ACC_ABSTRACT) printf("%sabstract", cnt++ ? " " : "");
}


/* Prints byte code. */
void printCode(clsDefinition *cls, unsigned char *p, jint len) {
	jint pc, wide = 0;
	for (pc = 0; pc < len; pc++) {
		printf("               %5u    ", pc);
		switch (p[pc]) {
			case 0:
				printf("nop\n"); 
				break;
			case 1:
				printf("aconst_null\n"); 
				break;
			case 2:
				printf("iconst_m1\n"); 
				break;
			case 3:
				printf("iconst_0\n"); 
				break;
			case 4:
				printf("iconst_1\n"); 
				break;
			case 5:
				printf("iconst_2\n"); 
				break;
			case 6:
				printf("iconst_3\n"); 
				break;
			case 7:
				printf("iconst_4\n"); 
				break;
			case 8:
				printf("iconst_5\n"); 
				break;
			case 9:
				printf("lconst_0\n"); 
				break;
			case 10:
				printf("lconst_1\n"); 
				break;
			case 11:
				printf("fconst_0\n"); 
				break;
			case 12:
				printf("fconst_1\n"); 
				break;
			case 13:
				printf("fconst_2\n"); 
				break;
			case 14:
				printf("dconst_0\n"); 
				break;
			case 15:
				printf("dconst_1\n"); 
				break;
			case 16:
				printf("bipush        %d\n", (jbyte)p[++pc]); 
				break;
			case 17:
				printf("sipush        %d\n", (jshort)((p[pc+1] << 8) | p[pc+2])); 
				pc += 2; 
				break;
			case 18: {
				jint idx = p[++pc];
				printf("ldc           #%u=%s\n", idx, getConstantString(cls, idx));
				break;
			}
			case 19: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("ldc_w         #%u=%s\n", idx, getConstantString(cls, idx)); 
				pc += 2; 
				break;
			}
			case 20: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("ldc2_w        #%u=%s\n", idx, getConstantString(cls, idx)); 
				pc += 2;
				break;
			}
			case 21:
				if (wide) {
					printf("iload         %u\n", (jushort)((p[pc+1] << 8) | p[pc+2]));
					pc += 2; wide = 0;					
				} else
					printf("iload         %u\n", p[++pc]);
				break;
			case 22:
				if (wide) {
					printf("lload         %u\n", (jushort)((p[pc+1] << 8) | p[pc+2]));
					pc += 2; wide = 0;					
				} else
					printf("lload         %u\n", p[++pc]);
				break;
			case 23:
				if (wide) {
					printf("fload         %u\n", (jushort)((p[pc+1] << 8) | p[pc+2]));
					pc += 2; wide = 0;					
				} else
					printf("fload         %u\n", p[++pc]);
				break;
			case 24:
				if (wide) {
					printf("dload         %u\n", (jushort)((p[pc+1] << 8) | p[pc+2]));
					pc += 2; wide = 0;					
				} else
					printf("dload         %u\n", p[++pc]);
				break;
			case 25:
				if (wide) {
					printf("aload         %u\n", (jushort)((p[pc+1] << 8) | p[pc+2]));
					pc += 2; wide = 0;					
				} else
					printf("aload         %u\n", p[++pc]);
				break;
			case 26:
				printf("iload_0\n"); 
				break;
			case 27:
				printf("iload_1\n"); 
				break;
			case 28:
				printf("iload_2\n"); 
				break;
			case 29:
				printf("iload_3\n"); 
				break;
			case 30:
				printf("lload_0\n"); 
				break;
			case 31:
				printf("lload_1\n"); 
				break;
			case 32:
				printf("lload_2\n"); 
				break;
			case 33:
				printf("lload_3\n"); 
				break;
			case 34:
				printf("fload_0\n"); 
				break;
			case 35:
				printf("fload_1\n"); 
				break;
			case 36:
				printf("fload_2\n"); 
				break;
			case 37:
				printf("fload_3\n"); 
				break;
			case 38:
				printf("dload_0\n"); 
				break;
			case 39:
				printf("dload_1\n"); 
				break;
			case 40:
				printf("dload_2\n"); 
				break;
			case 41:
				printf("dload_3\n"); 
				break;
			case 42:
				printf("aload_0\n"); 
				break;
			case 43:
				printf("aload_1\n"); 
				break;
			case 44:
				printf("aload_2\n"); 
				break;
			case 45:
				printf("aload_3\n");
				break;
			case 46:
				printf("iaload\n"); 
				break;
			case 47:
				printf("laload\n"); 
				break;
			case 48:
				printf("faload\n"); 
				break;
			case 49:
				printf("daload\n"); 
				break;
			case 50:
				printf("aaload\n"); 
				break;
			case 51:
				printf("baload\n"); 
				break;
			case 52:
				printf("caload\n"); 
				break;
			case 53:
				printf("saload\n"); 
				break;
			case 54:
				if (wide) {
					printf("istore        %u\n", (jushort)((p[pc+1] << 8) | p[pc+2]));
					pc += 2; wide = 0;					
				} else
					printf("istore        %u\n", p[++pc]);
				break;
			case 55:
				if (wide) {
					printf("lstore        %u\n", (jushort)((p[pc+1] << 8) | p[pc+2]));
					pc += 2; wide = 0;					
				} else
					printf("lstore        %u\n", p[++pc]);
				break;
			case 56:
				if (wide) {
					printf("fstore        %u\n", (jushort)((p[pc+1] << 8) | p[pc+2]));
					pc += 2; wide = 0;					
				} else
					printf("fstore        %u\n", p[++pc]);
				break;
			case 57:
				if (wide) {
					printf("dstore        %u\n", (jushort)((p[pc+1] << 8) | p[pc+2]));
					pc += 2; wide = 0;					
				} else
					printf("dstore        %u\n", p[++pc]);
				break;
			case 58:
				if (wide) {
					printf("astore        %u\n", (jushort)((p[pc+1] << 8) | p[pc+2]));
					pc += 2; wide = 0;					
				} else
					printf("astore        %u\n", p[++pc]);
				break;
			case 59:
				printf("istore_0\n"); 
				break;
			case 60:
				printf("istore_1\n"); 
				break;
			case 61:
				printf("istore_2\n"); 
				break;
			case 62:
				printf("istore_3\n"); 
				break;
			case 63:
				printf("lstore_0\n"); 
				break;
			case 64:
				printf("lstore_1\n"); 
				break;
			case 65:
				printf("lstore_2\n"); 
				break;
			case 66:
				printf("lstore_3\n"); 
				break;
			case 67:
				printf("fstore_0\n"); 
				break;
			case 68:
				printf("fstore_1\n"); 
				break;
			case 69:
				printf("fstore_2\n"); 
				break;
			case 70:
				printf("fstore_3\n"); 
				break;
			case 71:
				printf("dstore_0\n"); 
				break;
			case 72:
				printf("dstore_1\n"); 
				break;
			case 73:
				printf("dstore_2\n"); 
				break;
			case 74:
				printf("dstore_3\n"); 
				break;
			case 75:
				printf("astore_0\n"); 
				break;
			case 76:
				printf("astore_1\n"); 
				break;
			case 77:
				printf("astore_2\n"); 
				break;
			case 78:
				printf("astore_3\n");
				break;
			case 79:
				printf("iastore\n");
				break;
			case 80:
				printf("lastore\n");
				break;
			case 81:
				printf("fastore\n"); 
				break;
			case 82:
				printf("dastore\n"); 
				break;
			case 83:
				printf("aastore\n"); 
				break;
			case 84:
				printf("bastore\n"); 
				break;
			case 85:
				printf("castore\n"); 
				break;
			case 86:
				printf("sastore\n"); 
				break;
			case 87:
				printf("pop\n"); 
				break;
			case 88:
				printf("pop2\n"); 
				break;
			case 89:
				printf("dup\n"); 
				break;
			case 90:
				printf("dup_x1\n"); 
				break;
			case 91:
				printf("dup_x2\n"); 
				break;
			case 92:
				printf("dup2\n"); 
				break;
			case 93:
				printf("dup2_x1\n"); 
				break;
			case 94:
				printf("dup2_x2\n"); 
				break;
			case 95:
				printf("swap\n"); 
				break;
			case 96:
				printf("iadd\n"); 
				break;
			case 97:
				printf("ladd\n"); 
				break;
			case 98:
				printf("fadd\n"); 
				break;
			case 99:
				printf("dadd\n"); 
				break;
			case 100:
				printf("isub\n"); 
				break;
			case 101:
				printf("lsub\n"); 
				break;
			case 102:
				printf("fsub\n"); 
				break;
			case 103:
				printf("dsub\n"); 
				break;
			case 104:
				printf("imul\n"); 
				break;
			case 105:
				printf("lmul\n"); 
				break;
			case 106:
				printf("fmul\n"); 
				break;
			case 107:
				printf("dmul\n"); 
				break;
			case 108:
				printf("idiv\n"); 
				break;
			case 109:
				printf("ldiv\n"); 
				break;
			case 110:
				printf("fdiv\n"); 
				break;
			case 111:
				printf("ddiv\n"); 
				break;
			case 112:
				printf("irem\n"); 
				break;
			case 113:
				printf("lrem\n"); 
				break;
			case 114:
				printf("frem\n"); 
				break;
			case 115:
				printf("drem\n"); 
				break;
			case 116:
				printf("ineg\n"); 
				break;
			case 117:
				printf("lneg\n"); 
				break;
			case 118:
				printf("fneg\n"); 
				break;
			case 119:
				printf("dneg\n"); 
				break;
			case 120:
				printf("ishl\n"); 
				break;
			case 121:
				printf("lshl\n"); 
				break;
			case 122:
				printf("ishr\n"); 
				break;
			case 123:
				printf("lshr\n"); 
				break;
			case 124:
				printf("iushr\n"); 
				break;
			case 125:
				printf("lushr\n"); 
				break;
			case 126:
				printf("iand\n"); 
				break;
			case 127:
				printf("land\n"); 
				break;
			case 128:
				printf("ior\n"); 
				break;
			case 129:
				printf("lor\n"); 
				break;
			case 130:
				printf("ixor\n"); 
				break;
			case 131:
				printf("lxor\n"); 
				break;
			case 132:
				if (wide) {
					printf("iinc          %u, %d\n", (jushort)((p[pc+1] << 8) | p[pc+2]), (jshort)((p[pc+3] << 8) | p[pc+4]));
					pc += 4; wide = 0;					
				} else {
					printf("iinc          %u, %d\n", p[pc+1], (jbyte)p[pc+2]);
					pc += 2;
				}
				break;
			case 133:
				printf("i2l\n"); 
				break;
			case 134:
				printf("i2f\n"); 
				break;
			case 135:
				printf("i2d\n"); 
				break;
			case 136:
				printf("l2i\n"); 
				break;
			case 137:
				printf("l2f\n"); 
				break;
			case 138:
				printf("l2d\n"); 
				break;
			case 139:
				printf("f2i\n"); 
				break;
			case 140:
				printf("f2l\n"); 
				break;
			case 141:
				printf("f2d\n"); 
				break;
			case 142:
				printf("d2i\n"); 
				break;
			case 143:
				printf("d2l\n"); 
				break;
			case 144:
				printf("d2f\n"); 
				break;
			case 145:
				printf("i2b\n"); 
				break;
			case 146:
				printf("i2c\n"); 
				break;
			case 147:
				printf("i2s\n"); 
				break;
			case 148:
				printf("lcmp\n"); 
				break;
			case 149:
				printf("fcmpl\n"); 
				break;
			case 150:
				printf("fcmpg\n"); 
				break;
			case 151:
				printf("dcmpl\n"); 
				break;
			case 152:
				printf("dcmpg\n"); 
				break;
			case 153:
				printf("ifeq          %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 154:
				printf("ifne          %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 155:
				printf("iflt          %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 156:
				printf("ifge          %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 157:
				printf("ifgt          %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 158:
				printf("ifle          %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 159:
				printf("if_icmpeq     %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 160:
				printf("if_icmpne     %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 161:
				printf("if_icmplt     %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 162:
				printf("if_icmpge     %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 163:
				printf("if_icmpgt     %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 164:
				printf("if_icmple     %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2;
				break;
			case 165:
				printf("if_acmpeq     %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 166:
				printf("if_acmpne     %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 167:
				printf("goto          %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2])); 
				pc += 2; 
				break;
			case 168:
				printf("jsr           %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2;
				break;
			case 169:
				if (wide) {
					printf("ret           %u\n", (jushort)((p[pc+1] << 8) | p[pc+2]));
					pc += 2; wide = 0;					
				} else
					printf("ret           %u\n", p[++pc]);
				break;
			case 170: {
				jint i, n1, n2, o, s; s = pc++;
				while (p[pc] == 0) {
					if ((pc % 4) == 0) break;
					pc++;
				}
				printf("tableswitch   default: %d\n", s + (jint)((p[pc] << 24) | (p[pc+1] << 16) | (p[pc+2] << 8) | p[pc+3]));
				pc += 4;
				n1 = (jint)((p[pc] << 24) | (p[pc+1] << 16) | (p[pc+2] << 8) | p[pc+3]); pc += 4;
				n2 = (jint)((p[pc] << 24) | (p[pc+1] << 16) | (p[pc+2] << 8) | p[pc+3]); pc += 4;
				for (i = n1; i <= n2; i++) {
					o = (jint)((p[pc] << 24) | (p[pc+1] << 16) | (p[pc+2] << 8) | p[pc+3]); pc += 4;
					printf("                                      %d: %d\n", i, s + o);
				}
				pc--;
				break;
			}
			case 171: {
				jint i, n, m, o, s; s = pc++;
				while (p[pc] == 0) {
					if ((pc % 4) == 0) break;
					pc++;
				}
				printf("lookupswitch  default: %d\n", s + (jint)((p[pc] << 24) | (p[pc+1] << 16) | (p[pc+2] << 8) | p[pc+3]));
				pc += 4;
				n = (jint)((p[pc] << 24) | (p[pc+1] << 16) | (p[pc+2] << 8) | p[pc+3]);
				pc += 4;
				for (i = 0; i < n; i++) {
					m = (jint)((p[pc] << 24) | (p[pc+1] << 16) | (p[pc+2] << 8) | p[pc+3]); pc += 4;
					o = (jint)((p[pc] << 24) | (p[pc+1] << 16) | (p[pc+2] << 8) | p[pc+3]); pc += 4;
					printf("                                      match %d: %d\n", m, s + o);
				}
				pc--;
				break; 
			}
			case 172:
				printf("ireturn\n"); 
				break;
			case 173:
				printf("lreturn\n"); 
				break;
			case 174:
				printf("freturn\n"); 
				break;
			case 175:
				printf("dreturn\n"); 
				break;
			case 176:
				printf("areturn\n"); 
				break;
			case 177:
				printf("return\n"); 
				break;
			case 178: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("getstatic     #%u=%s\n", idx, getConstantString(cls, idx)); 
				pc += 2; 
				break;
			}
			case 179: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("putstatic     #%u=%s\n", idx, getConstantString(cls, idx)); 
				pc += 2; 
				break;
			}
			case 180: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("getfield      #%u=%s\n", idx, getConstantString(cls, idx)); 
				pc += 2; 
				break;
			}
			case 181: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("putfield      #%u=%s\n", idx, getConstantString(cls, idx)); 
				pc += 2; 
				break;
			}
			case 182: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("invokevirtual #%u=%s\n", idx, getConstantString(cls, idx)); 
				pc += 2; 
				break;
			}
			case 183: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("invokespecial #%u=%s\n", idx, getConstantString(cls, idx)); 
				pc += 2; 
				break;
			}
			case 184: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("invokestatic  #%u=%s\n", idx, getConstantString(cls, idx)); 
				pc += 2; 
				break;
			}
			case 185: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("invokeinterface #%u=%s, %u\n", idx, getConstantString(cls, idx), (unsigned char)p[pc+3]); 
				pc += 4; 
				break;
			}
			case 187: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("new           #%u=%s\n", idx, getConstantString(cls, idx)); 
				pc += 2; 
				break;
			}
			case 188: {
				char *t = "???";
				switch (p[++pc]) {
					case 4: t = "boolean"; break;
					case 5: t = "char"; break;
					case 6: t = "float"; break;
					case 7: t = "double"; break;
					case 8: t = "byte"; break;
					case 9: t = "short"; break;
					case 10: t = "int"; break;
					case 11: t = "long"; break;
				}
				printf("newarray      %s\n", t); 
				break;
			}
			case 189: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("anewarray     #%u=%s\n", idx, getConstantString(cls, idx)); 
				pc += 2; 
				break;
			}
			case 190:
				printf("arraylength\n"); 
				break;
			case 191:
				printf("athrow\n"); 
				break;
			case 192: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("checkcast     #%u=%s\n", idx, getConstantString(cls, idx));
				pc += 2; 
				break;
			}
			case 193: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("instanceof    #%u=%s\n", idx, getConstantString(cls, idx));
				pc += 2; 
				break;
			}
			case 194:
				printf("monitorenter\n"); 
				break;
			case 195:
				printf("monitorexit\n"); 
				break;
			case 196:
				printf("wide ");
				wide = 1;
				break;
			case 197: {
				jint idx = (jushort)((p[pc+1] << 8) | p[pc+2]);
				printf("multianewarray #%u=%s, %u\n", idx, getConstantString(cls, idx), p[pc+3]);
				pc += 3;
				break;
			}
			case 198:
				printf("ifnull        %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 199:
				printf("ifnonnull     %u\n", pc + (jshort)((p[pc+1] << 8) | p[pc+2]));
				pc += 2; 
				break;
			case 200:
				printf("goto_w        %u\n", (jint)((p[pc+1] << 24) | (p[pc+2] << 16) | (p[pc+3] << 8) | p[pc+4])); 
				pc += 4; 
				break;
			case 201:
				printf("jsr_w         %u\n", (jint)((p[pc+1] << 24) | (p[pc+2] << 16) | (p[pc+3] << 8) | p[pc+4])); 
				pc += 4; 
				break;
			default: 
				printf("unknown code: %02X\n", p[pc]);
		}
	}
}


/* Prints the attribute. */
void printAttribute(clsDefinition *cls, clsAttribute *atr, int i, const char* prf) {
	int j = 0, k;
	if (i < 0)
		i = -i;
	else
		atr += i;
	printf("%s%4u) %s\n", prf, i + 1, atr->name);
	if (atr <= 0) 
		return;
	
	/* SourceFile attribut? */
	if (exp_atr && !vmStrCmpUTF(atr->name, "SourceFile")) {
		printf("      %s\n", cls->constant_pool[getClassUShort(atr->data)].p);
		return;
	}

	/* InnerClasses attribut? */
	if (exp_atr && !vmStrCmpUTF(atr->name, "InnerClasses")) {
		jint i, count = getClassUShort(atr->data);
		for (i = 0; i < count; i++) {
			int InnerClass = getClassUShort(atr->data + 2 + 8 * i);
			int OuterClass = getClassUShort(atr->data + 4 + 8 * i);
			int InnerName = getClassUShort(atr->data + 6 + 8 * i);
			printf("%s      {", prf);
			printf("INNERCLASS=#%u%s%s, ", InnerClass, InnerClass ? "=" : "",
				InnerClass ? cls->constant_pool[InnerClass].p : "");
			printf("OUTERCLASS=#%u%s%s, ", OuterClass, OuterClass ? "=" : "",
				OuterClass ? cls->constant_pool[OuterClass].p : "");
			printf("INNERNAME=#%u%s%s, INNERFLAGS=", InnerName, InnerName ? "=" : "",
				InnerName ? cls->constant_pool[InnerName].p : "");
			printClassFlags(getClassUShort(atr->data + 8 + 8 * i));
			printf("}\n");
		}
		return;
	}

	/* Exceptions attribut? */
	if (exp_atr && !vmStrCmpUTF(atr->name, "Exceptions")) {
		jint i, count = getClassUShort(atr->data);
		for (i = 1; i <= count; i++) {
			int exp = getClassUShort(atr->data + 2 * i);
			printf("%s      %u) #%u%s%s\n", prf, i, exp, exp ? "=" : "",
				exp ? cls->constant_pool[exp].p : "");
		}
		return;
	}

	/* ConstantValue attribut? */
	if (exp_atr && !vmStrCmpUTF(atr->name, "ConstantValue")) {
		jint i = getClassUShort(atr->data);
		clsConstPool *c = cls->constant_pool + i;
		printf("%s      ", prf);
		switch (c->tag) {
			case CONSTANT_Integer:
				printf("%i", (int)*(jint*)(c->p));
				break;
			case CONSTANT_Long:
				printf("%s", getLong(*(jlong*)(c->p)));
				break;
			case CONSTANT_Float:
				printf("%g", (double)*(jfloat*)(c->p));
				break;
			case CONSTANT_Double:
				printf("%g", (double)*(jdouble*)(c->p));
				break;
			case CONSTANT_String:
				printf("#%u=", c->i);
				printUtf8(cls->constant_pool[c->i].p, cls->constant_pool[c->i].i);
				break;
		}
		printf("\n");
		return;
	}

	/* LineNumberTable attribut? */
	if (exp_atr && !vmStrCmpUTF(atr->name, "LineNumberTable")) {
		jint i, count = getClassUShort(atr->data);
		for (i = 0; i < count; i++)
			printf("%s      %3u) Pc=%u, Line=%u\n", prf, i + 1, 
				getClassUShort(atr->data + 2 + 4 * i),
				getClassUShort(atr->data + 4 + 4 * i));
		return;
	}

	/* LocalVariableTable attribut? */
	if (exp_atr && !vmStrCmpUTF(atr->name, "LocalVariableTable")) {
		jint i, count = getClassUShort(atr->data);
		for (i = 0; i < count; i++) {
			jbyte *p = atr->data + 2 + 10 * i;
			printf("%s      %3u) Pc=%u-%u, Name=#%u=%s, Descriptor=#%u=%s, Index=%u\n", 
				prf, i + 1, getClassUShort(p), getClassUShort(p) + getClassUShort(p+2),
				getClassUShort(p+4), cls->constant_pool[getClassUShort(p+4)].p,
				getClassUShort(p+6), cls->constant_pool[getClassUShort(p+6)].p,
				getClassUShort(p+8)
			);
		}
		return;
	}

	/* Code attribut? */
	if (exp_atr && !vmStrCmpUTF(atr->name, "Code")) {
		jint i, cnt; jbyte *p;
		printf("%s      Maximum Operand Stack:    %u\n", prf, getClassUShort(atr->data));
		printf("%s      Local variables count:    %u\n", prf, getClassUShort(atr->data + 2));
		cnt = getClassInt(atr->data + 4);
		printf("%s      Code (length=%u):\n", prf, cnt);
		printCode(cls, atr->data + 8, cnt);
		p = atr->data + 8 + cnt;
		cnt = getClassUShort(p); p += 2;
		if (cnt > 0) printf("%s      Exception table:\n", prf);
		for (i = 0; i < cnt; i++) {
			int exp = getClassUShort(p+6);
			printf("%s      %4u) Pc=%u-%u, HandlerPc=%u, Type=#%u%s%s\n", prf, i+1,
				getClassUShort(p), getClassUShort(p+2), getClassUShort(p+4),
				exp, exp ? "=" : "", exp ? cls->constant_pool[exp].p : ""
			);
			p += 8;
		}
		cnt = getClassUShort(p); p += 2;
		if (cnt > 0) printf("%s      Attributes:\n", prf);
		for (i = 0; i < cnt; i++) {
			clsAttribute atr;
			atr.name = cls->constant_pool[getClassUShort(p)].p;
			atr.len = getClassInt(p + 2);
			atr.data = p + 6;
			p += 6 + atr.len;
			printAttribute(cls, &atr, -i, "            ");
		}
		return;
	}

	/* binary data */
	while (j < atr->len) {
		char buf[256], pom[64]; int n; buf[0] = 0;
		for (k = j, n = 0; (k < atr->len) && (n < 8); k++, n++) {
			vmSprintfUTF(pom, "%02X ", (unsigned char)atr->data[k]);
			vmStrCatUTF(buf, pom);
		}
		vmStrCatUTF(buf, "  ");
		for (; (k < atr->len) && (n < 16); k++, n++) {
			vmSprintfUTF(pom, "%02X ", (unsigned char)atr->data[k]);
			vmStrCatUTF(buf, pom);
		}		
		printf("%s      %08X:  %s\n", prf, j, buf);
		j += 16; 
	}
}


/* Prints the constant. */
int printConstant(clsDefinition *cls, clsConstPool *cp, int i) {
	int ret = 1;
	clsConstPool *c = cp + i;
	printf("%4u) %s: ", i, cp_tags[c->tag]);
	switch (c->tag) {
		case CONSTANT_Utf8:
			printf("     ");
			printUtf8(c->p, c->i);
			break;
		case CONSTANT_Class:
			printf("    #%u=%s", c->i, cp[c->i].p);
			break;
		case CONSTANT_String:
			printf("   #%u=", c->i);
			printUtf8(cp[c->i].p, cp[c->i].i);
			break;
		case CONSTANT_Integer:
			printf("  %i", (int)*(jint*)(c->p));
			break;
		case CONSTANT_Float:
			printf("    %g", (double)*(jfloat*)(c->p));
			break;
		case CONSTANT_Long:
			printf("     %s", getLong(*(jlong*)(c->p)));
			ret = 2;
			break;
		case CONSTANT_Double: 
			printf("   %g", (double)*(jdouble*)(c->p));
			ret = 2;
			break;
		case CONSTANT_NameAndType: 
			printf("NAME=#%u=%s, DESCRIPTOR=#%u=%s", c->i, cp[c->i].p, c->j, cp[c->j].p);
			break;
		case CONSTANT_FieldRef:
			printf(" CLASS=#%u=%s, NAME&TYPE=#%u={NAME=#%u=%s, DESCRIPTOR=#%u=%s}", 
				c->i, cp[c->i].p, c->j, cp[c->j].i, cp[cp[c->j].i].p, cp[c->j].j, cp[cp[c->j].j].p);
			break;
		case CONSTANT_MethodRef: 
		case CONSTANT_InterfaceMethodRef:
			printf("CLASS=#%u=%s, NAME&TYPE=#%u={NAME=#%u=%s, DESCRIPTOR=#%u=%s}", 
				c->i, cp[c->i].p, c->j, cp[c->j].i, cp[cp[c->j].i].p, cp[c->j].j, cp[cp[c->j].j].p);
			break;
	};
	printf("\n");
	return ret;
}


/* Prints the field. */
void printField(clsDefinition *cls, clsField *fld, int i, int i2) {
	fld += i;

	printf("%4u) ", i2 + 1);
	if (fld->access_flags & ACC_PUBLIC) printf("public ");
	if (fld->access_flags & ACC_PRIVATE) printf("private ");
	if (fld->access_flags & ACC_PROTECTED) printf("protected ");
	if (fld->access_flags & ACC_STATIC) printf("static ");
	if (fld->access_flags & ACC_FINAL) printf("final ");
	if (fld->access_flags & ACC_VOLATILE) printf("volatile ");
	if (fld->access_flags & ACC_TRANSIENT) printf("transient ");
	printf("%s %s;\n", getType(fld->descriptor), fld->name);

	/* field atributes */
	if (fld->attributes_count > 0) {
		printf("      Attributes:\n");
		for (i = 0; i < fld->attributes_count; i++)
			printAttribute(cls, fld->attributes, i, "      ");
	}
}


/* Prints the method. */
void printMethod(clsDefinition *cls, clsMethod **mets, int i, int i2) {
	const char *p;
	clsMethod *met = mets[i];

	printf("%4u) ", i2 + 1);
	if (met->access_flags & ACC_PUBLIC) printf("public ");
	if (met->access_flags & ACC_PRIVATE) printf("private ");
	if (met->access_flags & ACC_PROTECTED) printf("protected ");
	if (met->access_flags & ACC_STATIC) printf("static ");
	if (met->access_flags & ACC_FINAL) printf("final ");
	if (met->access_flags & ACC_SYNCHRONIZED) printf("synchronized ");
	if (met->access_flags & ACC_NATIVE) printf("native ");
	if (met->access_flags & ACC_ABSTRACT) printf("abstract ");
	if (met->access_flags & ACC_STRICT) printf("strict ");
	p = met->descriptor;
	while (*p && (*p != ')')) p++; if (*p == ')') p++;
	printf("%s %s(", getType(p), met->name);
	p = met->descriptor;
	while (*p && (*p != '(')) p++;
	while (*p && (*p != ')')) {
		if (*p == '(') p++; else printf(", ");
		if (*p == ')') break;
		printf(getType(p));
		while (*p && (*p == '[')) p++;
		if (*p == 'L') {
			while (*p && (*p != ';')) p++;
			if (*p == ';') p++;
		} else
			if (*p) p++;
	}
	printf(");\n");

	/* method atributes */
	printf("      Attributes:\n");
	for (i = 0; i < met->attributes_count; i++)
		printAttribute(cls, met->attributes, i, "      ");
}


/* Prints the class. */
int printClass(clsDefinition *cls) {
	int i, j;

	/* class properties */
	printf("Class:               %s\n", cls->name);
	printf("Super Class:         %s\n", cls->super_class);
	printf("Version:             %u.%u\n", cls->major_version, cls->minor_version);
	printf("Flags:               ");
	printClassFlags(cls->access_flags);
	printf("\nAttributes Count:    %u\n", cls->attributes_count);
	printf("Constant Pool Count: %u\n", cls->constant_pool_count);
	printf("Fields Count:        %u\n", cls->class_fields_count + cls->instance_fields_count);
	printf("Methods Count:       %u\n", cls->class_methods_count + cls->instance_methods_count);
	
	/* class atributes */
	printf("\n\nATTRIBUTES:\n===========\n");
	for (i = 0; i < cls->attributes_count; i++)
		printAttribute(cls, cls->attributes, i, "");

	/* constant pool */
	printf("\n\nCONSTANT POOL:\n==============\n");
	for (i = 1; i < cls->constant_pool_count; )
		i += printConstant(cls, cls->constant_pool, i);

	/* fields */
	printf("\n\nFIELDS:\n=======\n");
	for (i = 0, j = 0; i < cls->class_fields_count; i++, j++)
		printField(cls, cls->class_fields, i, j);
	for (i = 0; i < cls->instance_fields_count; i++, j++)
		printField(cls, cls->instance_fields, i, j);

	/* methods */
	printf("\n\nMETHODS:\n========\n");
	for (i = 0, j = 0; i < cls->class_methods_count; i++, j++)
		printMethod(cls, cls->class_methods, i, j);
	for (i = 0; i < cls->instance_methods_count; i++, j++)
		printMethod(cls, cls->instance_methods, i, j);

	printf("\n");
	return 1;
}


/**
   The program entry point.
   @param argc the count of arguments that follow in @a argv
   @param argv an array of strings representing command-line arguments entered by the user of the program
   @return the program exit code (0 on success)
*/
int main(int argc, char* argv[]) {
	int ret;
	FILE *f;
	char *fname;               /* source .class file name */
	jbyte *buf;                /* buffer containing the .class file data */
	jsize buflen;              /* buffer length */
	clsDefinition *cls;        /* class definition */
	ret = -1; buf = NULL; buflen = 0; cls = NULL;
	fname = argv[1];

	/* display help? */
	if ((argc != 2) && (argc != 3)) {
		ret = 0;
PRINT_HELP:
		printf("DS Java Disassembler 1.02, Copyright (c) 2006, 2007 Dusan Stastny\n\n");
		if (ret == -1)
			printf("ERROR: File not found.\n\n");
		if (ret == -4)
			printf("ERROR: Class file corrupted.\n\n");
		if (ret == -5)
			printf("ERROR: Unknown option '%s'.\n\n", argv[1]);
		printf("Using:      jdis [-b] source\n\n");
		printf("  -b        not expand attributes\n");
		printf("  source    .class file\n\n");
		return ret;
	}
	
	if (argc == 3) {
		fname = argv[2];
		if (vmStrCmpiUTF(argv[1], "-b")) {
			ret = -5;
			goto PRINT_HELP;
		}
		exp_atr = 0;
	}

	/* load file content */
	f = fopen(fname, "rb");
	if (f != NULL) {
		if (0 == fseek(f, 0, SEEK_END)) {
			buflen = ftell(f);
			fseek(f, 0, SEEK_SET);
			buf = vmMemAlloc(buflen);
			if (buf != NULL)
				if (buflen != fread(buf, 1, buflen, f)) {
					vmMemFree(buf); 
					buf = NULL;
				}
		} else
			ret = -2;
		fclose(f);
	} else
		goto PRINT_HELP;

	/* get class definition structure */
	if (buf != 0)
		cls = vmLoadClass(buf, buflen);
	else
		goto PRINT_HELP;

	/* print class to output */
	ret = -4;
	if (cls != NULL)
		ret = printClass(cls);
	else
		goto PRINT_HELP;

	/* free memory */
	if (buf != NULL) vmMemFree(buf);
	return ret;
}

